#!/usr/bin/env python
# -*- coding: utf-8 -*- 

import sys
from twisted.python import log

class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

def start(where=sys.stdout):
    """
    Starts logging  
    
    Args:
        where: where to log (default=sys.stdout)
    """
    log.startLogging(where)
    
def debug(message):
    """
    Publish a message to the global log publisher.
    
    Args:
        message: the log message
    """
    log.msg(message)
    
def error(message, excpection=Exception):
    """
    Publish an error to the global log publisher.
    
    Args:
        message: the error message
        exception: the exception type (default=Exception)
    """
    log.err(exception, message)


#!/usr/bin/env python
# -*- coding: utf-8 -*- 

import os
import json

def load(scope='default'):
    """
    loads choosen config file
    if scope is not define loads default (first try `local`, then `dev` and last `prod`)
    
    Args:
        scope: config file, without `.json`
    """
    
    if scope is 'default':
        for scope in ['local', 'dev', 'prod']:
            path = 'config/{}.json'.format(scope)
            if os.path.exists(path):
                config = load(scope)
                if config:
                    return config
                    
    else:
        path = 'config/{}.json'.format(scope)
        if os.path.exists(path):
            with open(path, 'r') as config:
                obj = json.loads(config.read())
                obj['scope'] = scope

                #Default Values

                _fillDefaultValue(obj, 'log', False)

                _fillDefaultValue(obj, 'server', {})
                _fillDefaultValue(obj['server'], 'port', 8888)
                _fillDefaultValue(obj['server'], 'cors', False)

                return obj
    raise EnvironmentError('No config file found')

def _fillDefaultValue(obj, key, default):
    """
    Fill Default value if key does not exists on object

    Args:
        obj: object to fill with default value
        key: key that will be filled if inexistent
        default: default value to be filled
    """
    if key not in obj:
        obj[key] = default

#shortcut to current config
value = load()


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import argparse

def main():

    parser = argparse.ArgumentParser(description='Pypolyback - a restful api framework.')
    parser.add_argument('-s', '--start', '--serve', 
                        action='store_true', default=False, 
                        help='starts server')
    parser.add_argument('-i','--init',
                        action='store_true', default=False, 
                        help='initialize project folder')
    parser.add_argument('-b', '--build', 
                        action='store_true', default=False, 
                        help='build project [not yet implemented]')
    parser.add_argument('-t', '--test',
                        action='store_true', default=False, 
                        help='run automated tests [not yet implemented]')
    
    args = parser.parse_args()

    if args.start:
        from pypolyback import server
        server.start()
    elif args.init:
        from pypolyback import project
        project.init()
    else:
        parser.print_help()


#!/usr/bin/env python
# -*- coding: utf-8 -*- 

from twisted.internet import reactor

from pypolyback import log, routes, config

def start():
    """
    Starts server
    """
    
    #log
    if config.value['log']:
        log.start()
    else:
        #TODO: log.start() on sql
        pass
    
    #settings
    settings = {
        "debug": config.value['scope'] is not 'prod'
    }
    
    if 'mail' in config.value:
        settings['email_settings'] = config.value['mail']

    application = routes.prepare()
    
    reactor.listenTCP(config.value['server']['port'], application)
    
    #start
    reactor.run()


#!/usr/bin/env python
# -*- coding: utf-8 -*- 

import json, os

def init():
    """
    Initiate project folders with sample config and readme
    """

    #folders
    dirs = [
        'config',
        'endpoints',
        'utils'
    ]
    _createDirs(dirs)

    #configs
    _writeConfigFile('prod', {
        'log': False,
        'server': {
            'port': 80,
            'cors': False
        }
    })

    _writeConfigFile('dev', {
        'log': True,
        'server': {
            'port': 8888,
            'cors': "*"
        }
    })

    #readme
    with open('./README.md', 'w') as outfile:
        outfile.write(readme)
    
def _createDirs(directories):
    for directory in directories:
        if not os.path.exists('./' + directory):
            os.makedirs(directory)


def _writeConfigFile(file, obj):
    with open('./config/' + file + '.json', 'w') as outfile:
        json.dump(obj, outfile, indent=4)

readme = """# MY PROJECT

My project description

## LANGUAGE
[Python 2.7](https://docs.python.org/2/tutorial/index.html)

## LIBRARIES
* [pypolyback](https://github.com/seijihirao/pypolyback) - Backend restful framework

---

## RUNNING

```
$ pypolyback
```
"""


#!/usr/bin/env python
# -*- coding: utf-8 -*- 

from pypolyback import config, log

class _ApiObject(object):
    """
    Api properties that will be avaliable on endpoints
    
    Properties:
        vars: dictionary api custom variables
        config: current config object
        supported_methods: supported http rest methods
        debug: function to log message
        error: function to log error
        (will be added endpoint required util modules)
    """

    vars = {}
    
    config = config.value

    supported_methods = [
        'get', 
        'post',
        'put',
        'delete',
        'head',
        'options',
        'default'
    ]

    def debug(self, msg):
        log.debug(msg)
        
    def error(self, msg):
        log.error(msg)

def mount():
    return _ApiObject()


"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages

try:
   import pypandoc
   description = pypandoc.convert('README.md', 'rst')
except (IOError, ImportError):
   description = ''

setup(
    name='pypolyback',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version='0.12.4',

    description='A simple back-end rest framework in python using twisted and cyclone',
    long_description=description,

    # The project's main homepage.
    url='https://github.com/seijihirao/pypolyback',

    # Author details
    author='Rodrigo Seiji Piubeli Hirao',
    author_email='rodrigo.seiji.hirao@gmail.com',

    # Choose your license
    license='MIT',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Application Frameworks',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
    ],

    # What does your project relate to?
    keywords='backend rest web service',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(exclude=['endpoints', 'config', 'utils']),

    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    #   py_modules=["my_module"],

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=['Twisted', 'cyclone'],

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
        'console_scripts': [
            'pypolyback=pypolyback.commandline:main'
        ],
    },
)

#
# Build project
# # python2 setup.py sdist
#
# Create wheel
# # python2 setup.py bdist_wheel
#
# Upload to Pypi
# # twine upload dist/pypolyback-x.y.z-py2-none-any.whl
# # # x.y.z - is the actual version


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys
import imp
import json
import cyclone.web
from twisted.internet.defer import inlineCallbacks as async

from pypolyback import log, apiobject

global_api = apiobject.mount()

class DynamicHandler(cyclone.web.RequestHandler):
    """
    Dummy cycloce Web Handler that will be used as a mold for endpoints 
    
    Properties:
        api: api properties that will be avaliable on endpoints, documented on `apiobject.py`
        pypoly_utils_get: list of `get` functions in the called util, they will be called before each endpoint `get` request
        pypoly_utils_post: list of `post` functions in the called util, they will be called before each endpoint `post` request
        pypoly_utils_any: list of `any` functions in the called util, they will be called before each endpoint request
    """
    
    def set_default_headers(self):
        """
        Set http header behaviour

        Access-Control-Allow-Origin: everyone
        """
        self.set_header('Access-Control-Allow-Origin', self.api.config['server']['cors'])

    @property
    def params(self):
        """
        Gets argument in list of dicts

        The returned value is always unicode.
        
        Returns:
            value: parameter value. If the argument is not present, returns an empty list.
        """
        
        if self._params is None:
            try:
                self._params = json.loads(self.request.body)
            except:
                self._params = {}
                for key in self.request.arguments:
                    if len(self.request.arguments[key]) is 1:
                        self._params[key] = self.request.arguments[key][0]
                    else:
                        self._params[key] = self.request.arguments[key]
                                
        return self._params
    
    def send(self, result):
        """
        Transform dicts in json and writes them using `cyclone.web.RequestHandler.write` method
        """
        if isinstance(result, dict):
            self.write(json.dumps(result, separators=(',', ':')))
        elif result:
            self.write(result)
    
    def pypoly_get(req, api, *args, **kwargs):
        """
        Function executed on http get that will be overriden by `get` function on your endpoint 
        
        Args:
            req: self, cyclone.web.RequestHandler
            api: object mounted on `apiobject.py` and incremented with util modules
            *args: don't really know
            **kwargs: don't really know
        
        Raises:
            NotImplemented: if endpoint get is called and not implemented
        """
        raise NotImplemented
        
    @async
    def get(self, *args, **kwargs):
        """
        Function executed on http get
        
        Args:
            *args: don't really know
            **kwargs: don't really know 
        """
        if self.pypoly_get_implemented:
            for util_func in self.pypoly_utils_any:
                util_func(self, self.api, *args, **kwargs)
                
            for util_func in self.pypoly_utils_get:
                util_func(self, self.api, *args, **kwargs)
                
            yield self.pypoly_get(self.api, *args, **kwargs)
        else:
            yield self.default(*args, **kwargs)
    
    def pypoly_post(req, api, *args, **kwargs):
        """
        Function executed on http post that will be overriden by `post` function on your endpoint 
        
        Args:
            req: self, cyclone.web.RequestHandler
            api: object mounted on `apiobject.py` and incremented with util modules
            *args: don't really know
            **kwargs: don't really know
        
        Raises:
            NotImplemented: if endpoint post is called and not implemented
        """
        raise NotImplemented
        
    @async
    def post(self, *args, **kwargs):
        """
        Function executed on http post
        
        Args:
            *args: don't really know
            **kwargs: don't really know
        """
        if self.pypoly_post_implemented:
            for util_func in self.pypoly_utils_any:
                util_func(self, self.api, *args, **kwargs)
                
            for util_func in self.pypoly_utils_post:
                util_func(self, self.api, *args, **kwargs)
                
            yield self.pypoly_post(self.api, *args, **kwargs)
        else:
            yield self.default(*args, **kwargs)
    
    def pypoly_put(req, api, *args, **kwargs):
        """
        Function executed on http put that will be overriden by `put` function on your endpoint 
        
        Args:
            req: self, cyclone.web.RequestHandler
            api: object mounted on `apiobject.py` and incremented with util modules
            *args: don't really know
            **kwargs: don't really know
        
        Raises:
            NotImplemented: if endpoint post is called and not implemented
        """
        raise NotImplemented
        
    @async
    def put(self, *args, **kwargs):
        """
        Function executed on http put
        
        Args:
            *args: don't really know
            **kwargs: don't really know
        """
        if self.pypoly_put_implemented:
            for util_func in self.pypoly_utils_any:
                util_func(self, self.api, *args, **kwargs)
                
            for util_func in self.pypoly_utils_put:
                util_func(self, self.api, *args, **kwargs)
                
            yield self.pypoly_put(self.api, *args, **kwargs)
        else:
            yield self.default(*args, **kwargs)
    
    def pypoly_delete(req, api, *args, **kwargs):
        """
        Function executed on http delete that will be overriden by `delete` function on your endpoint 
        
        Args:
            req: self, cyclone.web.RequestHandler
            api: object mounted on `apiobject.py` and incremented with util modules
            *args: don't really know
            **kwargs: don't really know
        
        Raises:
            NotImplemented: if endpoint post is called and not implemented
        """
        raise NotImplemented
        
    @async
    def delete(self, *args, **kwargs):
        """
        Function executed on http delete
        
        Args:
            *args: don't really know
            **kwargs: don't really know
        """
        print self
        if self.pypoly_delete_implemented:
            for util_func in self.pypoly_utils_any:
                util_func(self, self.api, *args, **kwargs)
                
            for util_func in self.pypoly_utils_delete:
                util_func(self, self.api, *args, **kwargs)
                
            yield self.pypoly_delete(self.api, *args, **kwargs)
        else:
            yield self.default(*args, **kwargs)
    
    def pypoly_head(req, api, *args, **kwargs):
        """
        Function executed on http head that will be overriden by `head` function on your endpoint 
        
        Args:
            req: self, cyclone.web.RequestHandler
            api: object mounted on `apiobject.py` and incremented with util modules
            *args: don't really know
            **kwargs: don't really know
        
        Raises:
            NotImplemented: if endpoint post is called and not implemented
        """
        raise NotImplemented
        
    @async
    def head(self, *args, **kwargs):
        """
        Function executed on http head
        
        Args:
            *args: don't really know
            **kwargs: don't really know
        """
        if self.pypoly_head_implemented:
            for util_func in self.pypoly_utils_any:
                util_func(self, self.api, *args, **kwargs)
                
            for util_func in self.pypoly_utils_head:
                util_func(self, self.api, *args, **kwargs)
                
            yield self.pypoly_head(self.api, *args, **kwargs)
        else:
            yield self.default(*args, **kwargs)
    
    def pypoly_options(req, api, *args, **kwargs):
        """
        Function executed on http options that will be overriden by `options` function on your endpoint 
        
        Args:
            req: self, cyclone.web.RequestHandler
            api: object mounted on `apiobject.py` and incremented with util modules
            *args: don't really know
            **kwargs: don't really know
        
        Raises:
            NotImplemented: if endpoint post is called and not implemented
        """
        pass
        
    @async
    def options(self, *args, **kwargs):
        """
        Function executed on http options
        
        Args:
            *args: don't really know
            **kwargs: don't really know
        """
        if self.pypoly_options_implemented:
            for util_func in self.pypoly_utils_any:
                util_func(self, self.api, *args, **kwargs)
                
            for util_func in self.pypoly_utils_options:
                util_func(self, self.api, *args, **kwargs)
                
            yield self.pypoly_options(self.api, *args, **kwargs)
        else:
            yield self.default(*args, **kwargs)
    
    def pypoly_default(req, api, *args, **kwargs):
        """
        Function executed on http default that will be overriden by `default` function on your endpoint 
        
        Args:
            req: self, cyclone.web.RequestHandler
            api: object mounted on `apiobject.py` and incremented with util modules
            *args: don't really know
            **kwargs: don't really know
        
        Raises:
            NotImplemented: if endpoint post is called and not implemented
        """
        raise NotImplemented
        
    @async
    def default(self, *args, **kwargs):
        """
        Function executed if a request does not match any implemented methods
        
        Args:
            *args: don't really know
            **kwargs: don't really know
        """
        for util_func in self.pypoly_utils_any:
            util_func(self, self.api, *args, **kwargs)
            
        for util_func in self.pypoly_utils_default:
            util_func(self, self.api, *args, **kwargs)
            
        yield self.pypoly_default(self.api, *args, **kwargs)



def prepare():
    """
    Mount routers and add them to cyclone application 
    
    Returns:
        application (cyclone.web.Application): cyclone application with endpoint routers mounted 
    """
    
    file_paths = []
    
    #read all endpoint files
    for root, subdirs, files in os.walk('./endpoints'):
        for file in files:
            if os.path.splitext(file)[1] == '.py':
                file_paths += [{
                    'url': os.path.splitext(os.path.relpath(os.path.join(root, file), 'endpoints/'))[0],
                    'file': os.path.join(root, file)
                }]
    
    routes = []
    
    utils = {}
    
    #populate routes
    for file_path in file_paths:
        
        log.debug('Loading endpoint: [' + log.bcolors.HEADER + file_path['url'] + log.bcolors.ENDC + ']')
        
        file_module = imp.load_source(file_path['url'].replace('/', '-'), file_path['file'])

        supported_methods = global_api.supported_methods

        handler_props = {
            'async': async,
            
            '_params': None,
            
            'api': global_api
        }

        ##
        # INITIALIZING METHOD PROPERTIES
        #
        for method in supported_methods:
            handler_props['pypoly_' + method + '_implemented'] = False
            handler_props['pypoly_utils_' + method] = []

        handler_props['pypoly_utils_any'] = []
        

        ##
        # ADDING METHOD FUNCTIONS
        #
        for method in supported_methods:
            if hasattr(file_module, method):
                handler_props['pypoly_' + method] = getattr(file_module, method)
                handler_props['pypoly_' + method + '_implemented'] = True
        
        ##
        # ADDING UTILS TO API PARAM
        #
        if hasattr(file_module, 'utils'):
            for util in file_module.utils:
                if not util in utils:
                    utils[util] = imp.load_source(util, os.path.join('utils', util + '.py'))
                    utils[util].async =  async
                setattr(handler_props['api'], util, utils[util])
                
                #calls util init function 
                if hasattr(utils[util], 'init'):
                    utils[util].init(global_api)
                
                #adds rest functionst
                for method in supported_methods:
                    if hasattr(utils[util], method):
                        handler_props['pypoly_utils_' + method] += [getattr(utils[util], method)]

                #adds 'any' function
                if hasattr(utils[util], 'any'):
                    handler_props['pypoly_utils_any'] += [utils[util].any]
        
        Handler = type(file_path['url'].replace('/', '').replace('.', ''), (DynamicHandler, ), handler_props)
        
        routes += [('/' + file_path['url'], Handler)]
        
        #Logging loaded endpoints
        log.debug('Endpoint Loaded: [' + log.bcolors.OKBLUE + file_path['url'] + log.bcolors.ENDC + ']')
    
    #Logging loades utils
    for root, subdirs, files in os.walk('./utils'):
        for file in files:
            if os.path.splitext(file)[1] == '.py':
                util = os.path.splitext(file)[0]
                if util in utils:
                    log.debug('Util Loaded: [' + log.bcolors.OKBLUE + util + log.bcolors.ENDC + ']')
                else:
                    log.debug('Util not Loaded: [' + log.bcolors.WARNING + util + log.bcolors.ENDC + ']')
    
    return cyclone.web.Application(routes)